import java.lang.*;
import java.util.*;
import java.io.*;

public class Dataguide {
    static private int BUFFER_SIZE = 1024*8;
    private Vector _out_edges_lists = null; // edge lists
    private Vector _in_edges_lists = null; // edge lists
    private Vector _dg_nodes = null; // nodes in the dataguide
    private DataguideNode _root_dg_node = null; // root node in dataguide
    private int _num_dg_nodes = 0; // number of nodes in dataguide
    private String _index_directory = null;
    private Vector _distances_from_root = null; // Vector<Integer> distances of nodes from root
    private DataGraph _data_graph = null;
    private int _num_file_reads = 0; 

    public Dataguide(DataGraph data_graph_) {
	// init
	_dg_nodes = new Vector();
	_out_edges_lists = new Vector();
	_in_edges_lists = new Vector();
	_data_graph = data_graph_;
	_index_directory = data_graph_.getIndexDirectory();

	// build
	File dg_file = new File(data_graph_.getFileName()+".dg");
	if(dg_file.exists()) {
	    System.out.println("Using existing dg files");
	    createDataguide3(data_graph_);
	}
	else {
	    createDataguide2(data_graph_);
	    if(data_graph_.getWriteFlag()) {
		if(MyFrame._debug_mode > 0) {
		    System.out.println("\tWriting dataguide to file");
		}
		long start_time = System.currentTimeMillis();
		writeDGToFile();
		long end_time   = System.currentTimeMillis();
		if(MyFrame._debug_mode > 0) {
		    System.out.println("\tWrote dataguide to file; Time taken = "+(end_time-start_time) +" milliseconds");
		}
	    }
	    else {		
		if(MyFrame._debug_mode > 0) {
		    System.out.println("Not writing dataguide to file");
		}
	    }

	}

	// optimize
	_dg_nodes.trimToSize();
	_num_dg_nodes = _dg_nodes.size();
	if(MyFrame._debug_mode > 2) {
	    System.out.println("Number of nodes in dataguide = " + _num_dg_nodes);
	}
	_out_edges_lists.trimToSize();	
	_in_edges_lists.trimToSize();	

	for(int i=0; i<_num_dg_nodes; i++) {
	    Vector out_edges_list = (Vector)_out_edges_lists.elementAt(i); // Vector<DataguideNode>
	    out_edges_list.trimToSize();
	    Vector in_edges_list = (Vector)_in_edges_lists.elementAt(i); // Vector<DataguideNode>
	    in_edges_list.trimToSize();
	}
	
    }

    /**
     * function to flush the dataguide nodes to disk.
     */
    public void flush() {
	Enumeration e = _dg_nodes.elements();
	while(e.hasMoreElements()) {
	    DataguideNode curr_dgnode = (DataguideNode)(e.nextElement());
	    curr_dgnode.flush();
	}
    }

    /**
     * function that creates the dataguide from the data graph
     *
     * @param data_graph_ data graph
     */
    private void createDataguide2(DataGraph data_graph_) { // create dg (seql io)
	// create the dataguide nodes & edges
	_out_edges_lists = data_graph_.getDataguideOutEdges();
	_in_edges_lists = data_graph_.getDataguideInEdges();
	_dg_nodes = data_graph_.getDataguideNodes();
	_root_dg_node = data_graph_.getDataguideRoot(); // root node in dataguide
	_num_dg_nodes = _dg_nodes.size();
	
	_distances_from_root = new Vector(_num_dg_nodes); 
	Enumeration e = _dg_nodes.elements();
	while(e.hasMoreElements()) {
	    DataguideNode curr_dgnode = (DataguideNode)(e.nextElement());
	    _distances_from_root.addElement(new Integer(curr_dgnode.getDistance()));
	}
    }

    /**
     * function that creates the dataguide from data written to disk in a previous session.
     *
     * @param data_graph_ data graph
     */
    private void createDataguide3(DataGraph data_graph_) { // create dg from file
	_data_graph = data_graph_;
	if(MyFrame._debug_mode > 0) {
	    System.out.println("\tReading from dataguide file");
	}
	long start_time = System.currentTimeMillis();
	readDGFromFile(data_graph_.getFileName()+".dg");
	long end_time = System.currentTimeMillis();
	if(MyFrame._debug_mode > 0) {
	    System.out.println("\tRead from dataguide file; Time taken = "+ (end_time-start_time) + " milliseconds");
	}
	
	_distances_from_root = new Vector(_num_dg_nodes); 
	Enumeration e = _dg_nodes.elements();
	while(e.hasMoreElements()) {
	    DataguideNode curr_dgnode = (DataguideNode)(e.nextElement());
	    _distances_from_root.addElement(new Integer(curr_dgnode.getDistance()));
	}
    }

    /**
     * function that writes the dataguide to disk.
     */
    void writeDGToFile() {
	// write dataguide to disk
	try {
	    String dg_file_name = _data_graph.getFileName()+".dg";
	    DataOutputStream dg_output_file = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(dg_file_name), BUFFER_SIZE));
	    dg_output_file.writeInt(_dg_nodes.size());
	    Enumeration e=_dg_nodes.elements();
	    while(e.hasMoreElements()) {
		DataguideNode curr_dgnode = (DataguideNode)(e.nextElement());
		dg_output_file.writeUTF(curr_dgnode.getLabel());
		dg_output_file.writeUTF(curr_dgnode.getPath());
		dg_output_file.writeInt(curr_dgnode.getDistance());

		int num_data_nodes = curr_dgnode.getNumDataNodes(); 
		dg_output_file.writeInt(num_data_nodes);
	    }
	    
	    dg_output_file.writeInt(_dg_nodes.indexOf(_root_dg_node));
	    
	    for(int i=0; i<_num_dg_nodes; i++) {
		Vector out_edges_list = (Vector)(_out_edges_lists.elementAt(i));
		int num_out_edges = out_edges_list.size();
		dg_output_file.writeInt(num_out_edges);
		if(MyFrame._debug_mode > 5) {
		    System.out.println(i+": number of out edges = "+num_out_edges);
		}
		for(int j=0; j<num_out_edges; j++) {
		    DataguideNode out_node = (DataguideNode)(out_edges_list.elementAt(j));
		    int out_edge = _dg_nodes.indexOf(out_node);
		    //System.out.println(i+": out edge to "+out_edge);
		    dg_output_file.writeInt(out_edge);
		}

		Vector in_edges_list = (Vector)(_in_edges_lists.elementAt(i));
		int num_in_edges = in_edges_list.size();
		dg_output_file.writeInt(num_in_edges);
		if(MyFrame._debug_mode > 5) {
		    System.out.println(i+": number of in edges = "+num_in_edges);
		}
		for(int j=0; j<num_in_edges; j++) {
		    DataguideNode in_node = (DataguideNode)(in_edges_list.elementAt(j));
		    int in_edge = _dg_nodes.indexOf(in_node);
		    //System.out.println(j+": in edge from "+in_edge);
		    dg_output_file.writeInt(in_edge);
		}
	    }

	    dg_output_file.close();
	    dg_output_file = null;
	}
	catch(IOException ioe) {
	    System.out.println("ioexception while writing dataguide " + ioe.toString());
	    System.exit(0);
	}
    }

    /**
     * function that reads a dataguide from disk.
     *
     * @param dg_file_name_ name of file
     */
    private void readDGFromFile(String dg_file_name_) {
	DataInputStream dg_input_file = null;
	try {
	    dg_input_file = new DataInputStream(new BufferedInputStream(new FileInputStream(dg_file_name_), BUFFER_SIZE));
	}
	catch(FileNotFoundException fnfe) {
	    System.out.println("File not found, file = " + dg_file_name_ + " "+ fnfe.toString());
	}
	try {
	    Random dg_node_id_gen = new Random();
	    _num_dg_nodes = dg_input_file.readInt();
	    System.out.println("Number of dg nodes = "+_num_dg_nodes);
	    _dg_nodes = new Vector();
	    for(int i=0; i<_num_dg_nodes; i++) {
		String curr_id = new String("Schema." + dg_node_id_gen.nextInt());
		String curr_label = dg_input_file.readUTF();
		String curr_path = dg_input_file.readUTF();
		int curr_dist = dg_input_file.readInt();
		int data_set_size = dg_input_file.readInt();
		if(MyFrame._debug_mode > 5) {
		    System.out.println("(l:p:d:s) = "+curr_label+":"+curr_path+":"+curr_dist+":"+data_set_size);
		}
		System.out.println(i+". "+curr_path);
		DataguideNode curr_dg_node = new DataguideNode(i, _index_directory, curr_id, curr_label);
		curr_dg_node.setPath(curr_path, curr_dist);
		curr_dg_node.setNumDataNodes(data_set_size);
		_dg_nodes.addElement(curr_dg_node);		
	    }
	    int root_index = dg_input_file.readInt();
	    _root_dg_node = (DataguideNode)(_dg_nodes.elementAt(root_index));	    

	    _out_edges_lists = new Vector();
	    _in_edges_lists = new Vector();
	    for(int i=0; i<_num_dg_nodes; i++) {		
		int num_out_edges = dg_input_file.readInt();
		if(MyFrame._debug_mode > 5) {
		    System.out.println(i+": number of out edges = "+num_out_edges);
		}
		Vector out_edges_list = new Vector(num_out_edges);
		for(int j=0; j<num_out_edges; j++) {		    
		    int out_edge = dg_input_file.readInt();
		    //System.out.println(i+": out edge to "+out_edge);
		    out_edges_list.addElement(_dg_nodes.elementAt(out_edge));
		}
		_out_edges_lists.addElement(out_edges_list);

		int num_in_edges = dg_input_file.readInt();
		if(MyFrame._debug_mode > 5) {
		    System.out.println(i+": number of in edges = "+num_in_edges);
		}
		Vector in_edges_list = new Vector(num_in_edges);
		for(int j=0; j<num_in_edges; j++) {
		    int in_edge = dg_input_file.readInt();
		    //System.out.println(j+": in edge from "+in_edge);
		    in_edges_list.addElement(_dg_nodes.elementAt(in_edge));
		}
		_in_edges_lists.addElement(in_edges_list);
	    }

	    dg_input_file.close();
	    dg_input_file = null;
	}
	catch(IOException ioe) {
	    System.out.println("ioexception during readDGFromFile" + ioe.toString());
	    System.exit(0);
	}
	
    }

    private void addEdgeBetween(DataguideNode src_dg_node_, DataguideNode dest_dg_node_) {
	// add dest_dg_node_ to the out list of src_dg_node_
	boolean src_found = false;
	int num_dg_nodes = _dg_nodes.size();
	int src_index=0;
	for(int i=0; i<num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)_dg_nodes.elementAt(i);
	    if(curr_dg_node == src_dg_node_) {src_found = true; src_index = i; break;}
	}
	if(src_found == false) {
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Dataguide node not found in dataguide"); // debug
		System.out.println("This must not have happened");
	    }
	    return;
	}
	Vector out_edges_list = (Vector)_out_edges_lists.elementAt(src_index);

	out_edges_list.addElement(dest_dg_node_);

	// add src_dg_node_ to the in list of dest_dg_node_
	boolean dst_found = false;
	int dst_index=0;
	for(int i=0; i<num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)_dg_nodes.elementAt(i);
	    if(curr_dg_node == dest_dg_node_) {dst_found = true; dst_index = i; break;}
	}
	if(dst_found == false) {
	    if(MyFrame._debug_mode > 5) {
		System.out.println("Dataguide node not found in dataguide"); // debug
		System.out.println("This must not have happened");
	    }
	    return;
	}
	Vector in_edges_list = (Vector)_in_edges_lists.elementAt(dst_index);
	in_edges_list.addElement(src_dg_node_);

	if(MyFrame._debug_mode > 7) {
	    System.out.println("Added dataguide edge " + src_dg_node_.getId() + 
			       " --> " + dest_dg_node_.getId());
	}
    }
    
    private boolean edgeExists(DataguideNode query_src_dg_node_, 
			       DataguideNode query_dest_dg_node_) {
	// find out_edge_list for query_src_dg_node_
	boolean src_found = false;
	int num_dg_nodes = _dg_nodes.size();
	int src_index = 0;
	for(int i=0; i<num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)_dg_nodes.elementAt(i);
	    if(curr_dg_node == query_src_dg_node_) {src_found = true; src_index=i; break;}
	}
	if(src_found == false) {
	    if(MyFrame._debug_mode > 0) {
		System.out.println("src dataguide node not found in dataguide"); 
		System.out.println("This must not have happened");
	    }
	    return false;
	}
	Vector out_edges_list = (Vector)_out_edges_lists.elementAt(src_index);
	
	// check if edge exists from query_src_dg_node_ to query_dest_dg_node_
	int num_out_edges = out_edges_list.size();
	for(int i=0; i<num_out_edges; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)out_edges_list.elementAt(i);
	    if(query_dest_dg_node_ == curr_dg_node) return true;
	}
	if(MyFrame._debug_mode > 7) {
	    System.out.println("edge (" + query_src_dg_node_.getId()+ "-->" + 
			       query_dest_dg_node_.getId() + ") not found in dataguide"); // debug
	}
	return false;
    }

    // returns the index of a dataguide node
    public int getNodeIndex(DataguideNode query_dg_node_) {
	for(int i=0; i<_num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)_dg_nodes.elementAt(i);
	    if(curr_dg_node == query_dg_node_) return i;
	}
	return -1;
    }

    /** 
     * function that returns the number of dg nodes.
     *
     * @return number of dg nodes
     */
    // return number of nodes in dataguide
    public int getNumNodes() {
	return _num_dg_nodes;
    }

    /**
     * function that returns the nodes connected to a given node.
     *
     * @param query_dg_node_ a dg node 
     * @return Vector containing all dg nodes connected to query_dg_node_
     */
    // return dg nodes that are linked to argument node
    public Vector getOutEdges(DataguideNode query_dg_node_) { // returns Vector<DataguideNode>
	// find the list of outgoing edges for query_dg_node_
	boolean found = false;
	int num_dg_nodes = _dg_nodes.size();
	int index=0;
	for(int i=0; i<num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)_dg_nodes.elementAt(i);
	    if(curr_dg_node == query_dg_node_) {found = true; index = i; break;}
	}
	if(found == false) {
	    if(MyFrame._debug_mode > 5) {
		System.out.println("dg node not found in dataguide"); 
	    }
	    return new Vector();
	}
	Vector out_edges_list = (Vector)_out_edges_lists.elementAt(index);
	return out_edges_list;
    }

    /**
     * function that returns the nodes the connect to a given node.
     *
     * @param query_dg_node_ a dg node
     * @return Vector containing all dg nodes that have an edge to query_dg_node_
     */
    public Vector getInEdges(DataguideNode query_dg_node_) { // returns Vector<DataguideNode>
	// find the list of incoming edges for query_dg_node_
	boolean found = false;
	int num_dg_nodes = _dg_nodes.size();
	int index=0;
	for(int i=0; i<num_dg_nodes; i++) {
	    DataguideNode curr_dg_node = (DataguideNode)_dg_nodes.elementAt(i);
	    if(curr_dg_node == query_dg_node_) {found = true; index = i; break;}
	}
	if(found == false) {
	    if(MyFrame._debug_mode > 5) {
		System.out.println("dg node not found in dataguide"); 
	    }
	    return new Vector();
	}
	Vector in_edges_list = (Vector)_in_edges_lists.elementAt(index);
	return in_edges_list;
    }

    /**
     * function that returns the root of the dataguide.
     *
     * @return DataguideNode object representing the root of the dataguide
     */
    // return root node of dataguide
    public DataguideNode getRoot() {
	return _root_dg_node; 
    }
    
    /**
     * function that returns all dg nodes at a given depth.
     *
     * @param depth_ depth
     * @return Vector of all dg nodes at depth = depth_
     */
    public Vector getNodesAtDepth(int depth_) { // root is at depth 0
	Vector nodes_at_depth = new Vector(); // Vector<DataguideNode>
	for(int i=0; i<_num_dg_nodes; i++) {
	    if(((Integer)_distances_from_root.elementAt(i)).intValue() == depth_) 
		nodes_at_depth.addElement(_dg_nodes.elementAt(i));
	}
	if(MyFrame._debug_mode > 5) {
	    System.out.println("Dataguide Nodes at depth " + depth_ + " = " + nodes_at_depth.size());
	}
	return nodes_at_depth;
    }

    /**
     * function that returns the data graph underlying this dataguide.
     * 
     * @return the data graph underlying this dataguide
     */
    public DataGraph getDataGraph() {
	return _data_graph;
    }


    /**
     * function that returns a dg node given its index.
     *
     * @param index_ index
     * @return DataguideNode object having the given index
     */
    public DataguideNode getDataguideNode(int index_) {
	if(index_ < _dg_nodes.size()) 
	    return (DataguideNode)(_dg_nodes.elementAt(index_));
	else
	    return null;
    }
}
