import java.lang.*;
import java.util.*;
import java.awt.geom.*;
import java.awt.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.*;

public class Histogram {
    static private int DEFAULT_NUM_BUCKETS = 5;
    private double _camera_mag = 0;
    private double _max_value = 0;
    private double _min_value = 0;

    private int _num_buckets = DEFAULT_NUM_BUCKETS;
    private DataguideNode _dgnode = null;

    private double _bucketWidth = 0;
    private Vector _buckets = null;
    private int _max_bucket_size = 0;
    private boolean _processing_complete = false;
    private int _batch_num = 0;
    private int _max_outliers = 0;
    private int _min_outliers = 0;

    static public  final double ABS_PEN_WIDTH = .1f;
    static public  final int MAX_BUCKETS = 200;
    static public  final double SPACING_X = 2.0f;
    static public  final int BATCH_SIZE = 100;
    static public  final double EPSILON = .01;

    /**
     * constructor.
     *
     * @param dgnode_ daaguide node that this hidtogram represents
     */
    Histogram(DataguideNode dgnode_) {
	_dgnode = dgnode_;
    }

    
    /**
     * function that returns the dataguide node that this histogram represents.
     * 
     * @return dataguide node that the current histogram object represents
     */
    DataguideNode getDGNode() {
	return _dgnode;
    }

    /**
     * function to set the number of buckets.
     *
     * @param num_ number of buckets
     */
    public void setNumBuckets(int num_) {
	_num_buckets = ((num_ < MAX_BUCKETS) ? num_ : MAX_BUCKETS);
    }
    
    /**
     * function that returns the number of buckets.
     *
     * @return number of buckets
     */
    public int getNumBuckets() {
	return _num_buckets;
    }

    /**
     * function that checks if processing for this histogram is complete.
     *
     * @return boolean value indicating if processing for this histogram is complete
     */
    public boolean processingComplete() {
	return _processing_complete;
    }

    /**
     * function that returns the number of the next batch to be processed.
     *
     * @return the number of the next batch to be processed
     */
    public int getNextBatchNumber() {
	_batch_num++;
	return _batch_num;
    }

    /**
     * function to find the bucket for a given data value.
     *
     * @param data_value_ the double whose bucket number is to be found
     * @return the number of the bucket data_value_ belongs to
     */
    private int findIndex(double data_value_) {
	double upper_range = _min_value + _bucketWidth;	
	int index = 0;
	while(data_value_ >= upper_range) {
	    upper_range = upper_range+_bucketWidth;
	    index++;
	}
	if(index >= _num_buckets) {
	    index = _num_buckets-1;
	}
	return index;
    }

    /**
     * function to compute the buckets for a vector of data values.
     *
     * @param datagraph_ data graph
     * @param data_ vector of data values
     */
    public void index(DataGraph datagraph_, FileVector data_, int batch_num_) {
	int num_data_items = data_.size();

	int start_index = batch_num_*BATCH_SIZE;
	if(batch_num_ == 0) {
	    if(num_data_items < _num_buckets) _num_buckets = num_data_items;	
	    boolean first_time = true;
	    
	    // read offsets into an array --- Thomas
	    for(int i=start_index; (i < (start_index+BATCH_SIZE)) && (i<num_data_items); i++) {
		//System.out.println("fmax: Getting offset "+i);
		long offset = data_.elementAt(i).longValue();
		DataNode curr_node = datagraph_.getDataNodeNoChildren(offset);
		String curr_data = curr_node.getData();
		double data_item = Double.parseDouble(curr_data);
		if(first_time) {
		    _max_value = data_item;
		    _min_value = data_item;
		    first_time = false;
		}
		else {
		    if(data_item > _max_value) _max_value = data_item;
		    if(data_item < _min_value) _min_value = data_item;
		}
		
	    }
	    	    
	    if((_max_value-_min_value) < EPSILON) {
		_num_buckets = 1;
		_bucketWidth = EPSILON;
	    }
	    else {
		_bucketWidth = (_max_value-_min_value)/_num_buckets;
	    }
	    _buckets = new Vector();
	    for(int j=0; j<_num_buckets; j++) {
		Integer bucket = new Integer(0);
		_buckets.addElement(bucket);
	    }
	    _processing_complete = false;
	    _max_bucket_size = 0;
	    _batch_num = 0;
	    _max_outliers = 0;
	    _min_outliers = 0;
	}

	for(int i=start_index; (i < (start_index+BATCH_SIZE)) && (i<num_data_items); i++) {
	    //System.out.println("index: Getting offset "+i);
	    long offset = data_.elementAt(i).longValue();
	    DataNode curr_node = datagraph_.getDataNodeNoChildren(offset);
	    //System.out.println("got node no children");
	    String curr_data = curr_node.getData();
	    //System.out.println("go data = "+curr_data);	    
	    try {
		double data_item = Double.parseDouble(curr_data);
		if(batch_num_ > 0 && data_item > _max_value) {_max_outliers++;}
		else if(batch_num_ > 0 && data_item < _min_value) {_min_outliers++;}
		else {
		    //System.out.println("parsed as double " + data_item);
		    int curr_index = this.findIndex(data_item);
		    //System.out.println("find index = "+curr_index);
		    Integer curr_bucket = (Integer)(_buckets.elementAt(curr_index));
		    int curr_count = curr_bucket.intValue();
		    curr_count++;
		    _buckets.setElementAt(new Integer(curr_count), curr_index);
		    if(curr_count > _max_bucket_size) _max_bucket_size = curr_count;
		}
	    }
	    catch(NumberFormatException nfe) {
		System.out.println("Histogram index: batch = "+batch_num_ +", index = "+i);
	    }
	    
	    if(i == (num_data_items - 1)) _processing_complete = true;
	}
	if(batch_num_ > 0) {
	    System.out.println("Outliers: max = "+_max_outliers +", min = "+_min_outliers);
	}
	if(!_processing_complete) {
	    System.out.println("Processed = "+((batch_num_+1)*BATCH_SIZE)+"/"+num_data_items);
	}
    }
    
    /**
     * function to compute the vector of Jazz visual components that represent this histogram.
     *
     * @param bounds_rect_ bounding rectangle for the histogram
     * @param camera_mag_ current magnification factor
     * @param colors_ Vector of colors available for drawing the histogram
     * @param avg_char_height_ average height of a character at the current magnification level
     * @param font_size_ integer font size for readable text
     * @return Vector of Jazz visual components that represent this histogram
     */
    public Vector display(Rectangle2D bounds_rect_, double camera_mag_, Vector colors_, double avg_char_height_, 
			  int font_size_) {
	Vector visual_components = new Vector();

	// labels
	int num_labels = (int)((bounds_rect_.getHeight()/avg_char_height_)/4.0);	
	if(_max_bucket_size < num_labels) num_labels = _max_bucket_size;
	double max_label_width = -1;
	int label_separation = 0;
	int curr_label = label_separation;
	
	if(num_labels > 0) {
	    label_separation = _max_bucket_size/num_labels;
	    for(int i=1; i<=num_labels; i++) {
		String curr_label_string = Integer.toString(curr_label);
		Font label_font = new Font(null, Font.PLAIN, font_size_);
		ZLabel label_visual_obj = new ZLabel(curr_label_string, label_font);
		label_visual_obj.setFont(label_font);
		label_visual_obj.setPenColor(Color.black);
		ZBounds bounds = label_visual_obj.getBounds();
		if(bounds.getWidth() > max_label_width) max_label_width = bounds.getWidth();
		curr_label += label_separation;
	    }

	    max_label_width += (2*SPACING_X/camera_mag_);
	    bounds_rect_.setRect(bounds_rect_.getX()+max_label_width, bounds_rect_.getY()+avg_char_height_/2.0f,
				 bounds_rect_.getWidth()-max_label_width, bounds_rect_.getHeight()-avg_char_height_/2.0f);
	}

	// vertical axis
	ZPolyline vert_axis = new ZPolyline();
	vert_axis.add(bounds_rect_.getX(), bounds_rect_.getY());
	vert_axis.add(bounds_rect_.getX(), bounds_rect_.getY()+bounds_rect_.getHeight());
	vert_axis.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	vert_axis.setPenColor(Color.black);
	visual_components.addElement(vert_axis);
			
	// horizontal axis
	ZPolyline hori_axis = new ZPolyline();
	hori_axis.add(bounds_rect_.getX(), bounds_rect_.getY()+bounds_rect_.getHeight());
	hori_axis.add(bounds_rect_.getX()+bounds_rect_.getWidth(), bounds_rect_.getY()+bounds_rect_.getHeight());
	hori_axis.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	vert_axis.setPenColor(Color.black);
	visual_components.addElement(hori_axis);
	
	// buckets
	double vbucket_width = bounds_rect_.getWidth()/_num_buckets;
	int num_colors = colors_.size();
	Enumeration e = _buckets.elements();
	int bkt_num = 0;
	double vbucket_bottomx = bounds_rect_.getX();
	double vbucket_bottomy = bounds_rect_.getY()+bounds_rect_.getHeight();
	while(e.hasMoreElements()) {
	    Integer bucket_size = (Integer)(e.nextElement());
	    double vbucket_height = (bounds_rect_.getHeight()/_max_bucket_size)*bucket_size.intValue();
	    Color vbucket_color = (Color)(colors_.elementAt(bucket_size.intValue()%num_colors));
	    ZRectangle vbucket = new ZRectangle(vbucket_bottomx, vbucket_bottomy-vbucket_height, 
						vbucket_width, vbucket_height);
	    vbucket.setPenColor(Color.black);
	    vbucket.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	    vbucket.setFillColor(vbucket_color);
	    visual_components.addElement(vbucket);
	    vbucket_bottomx += vbucket_width;
	    bkt_num++;
	}

	if(num_labels > 0) {
	    // labels
	    double height_separation = (bounds_rect_.getHeight()/_max_bucket_size)*(label_separation);
	    curr_label = label_separation;
	    for(int i=1; i<=num_labels; i++) {
		String curr_label_string = Integer.toString(curr_label);
		Font label_font = new Font(null, Font.PLAIN, font_size_);
		ZLabel label_visual_obj = new ZLabel(curr_label_string, label_font);
		label_visual_obj.setFont(label_font);
		label_visual_obj.setPenColor(Color.black);
		ZBounds bounds = label_visual_obj.getBounds();
		Point2D location = new Point2D.Double(bounds_rect_.getX()-bounds.getWidth()-SPACING_X/camera_mag_, 
						      bounds_rect_.getY()+ bounds_rect_.getHeight()
						      -i*height_separation
						      -avg_char_height_/2.0);
		label_visual_obj.setTranslation(location);
		visual_components.addElement(label_visual_obj);
		
		ZPolyline mark = new ZPolyline();
		mark.add(bounds_rect_.getX()-SPACING_X/(camera_mag_*2.0), bounds_rect_.getY()+ bounds_rect_.getHeight()-i*height_separation);
		mark.add(bounds_rect_.getX()+SPACING_X/(camera_mag_*2.0), bounds_rect_.getY()+ bounds_rect_.getHeight()-i*height_separation);
		mark.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
		mark.setPenColor(Color.black);
		visual_components.addElement(mark);
		
		curr_label += label_separation;
	    }
	}

	// legend
	
	String max_string = "Max. Val="+Double.toString(_max_value);
	String min_string = "Min. Val="+Double.toString(_min_value);
	int bucketWidth_int = (int)(_bucketWidth);
	double bucketWidth_fra = (int)((_bucketWidth-bucketWidth_int)*100)/(double)100;
	double bucketWidth = bucketWidth_int + bucketWidth_fra;
	String bkt_width  = "Bkt. Width = "+Double.toString(bucketWidth);
	Font label_font = new Font(null, Font.PLAIN, font_size_);
	double top_boundary = bounds_rect_.getY() + SPACING_X/camera_mag_;

	ZLabel maxstring_visual_obj = new ZLabel(max_string, label_font);
	maxstring_visual_obj.setFont(label_font);
	maxstring_visual_obj.setPenColor(Color.black);
	ZBounds maxstring_bounds = maxstring_visual_obj.getBounds();
	if(maxstring_bounds.getWidth() < bounds_rect_.getWidth() &&
	   ((top_boundary+maxstring_bounds.getHeight()+SPACING_X/camera_mag_) < 
	    (bounds_rect_.getY()+bounds_rect_.getWidth()))) {
	    Point2D maxstring_location = new Point2D.Double(bounds_rect_.getX()+bounds_rect_.getWidth()/2.0
							    -maxstring_bounds.getWidth()/2.0f, top_boundary);
	    maxstring_visual_obj.setTranslation(maxstring_location);
	    visual_components.addElement(maxstring_visual_obj);	
	    top_boundary += (maxstring_bounds.getHeight()+SPACING_X/camera_mag_);
	}

	ZLabel minstring_visual_obj = new ZLabel(min_string, label_font);
	minstring_visual_obj.setFont(label_font);
	minstring_visual_obj.setPenColor(Color.black);
	ZBounds minstring_bounds = minstring_visual_obj.getBounds();
	if(minstring_bounds.getWidth() < bounds_rect_.getWidth() &&
	   ((top_boundary+minstring_bounds.getHeight()+SPACING_X/camera_mag_) < 
	    (bounds_rect_.getY()+bounds_rect_.getWidth()))) {
	    Point2D minstring_location = new Point2D.Double(bounds_rect_.getX()+bounds_rect_.getWidth()/2.0
							    -minstring_bounds.getWidth()/2.0f, top_boundary);
	    minstring_visual_obj.setTranslation(minstring_location);
	    visual_components.addElement(minstring_visual_obj);	
	    top_boundary += (minstring_bounds.getHeight()+SPACING_X/camera_mag_);
	}

	ZLabel bktwidth_visual_obj = new ZLabel(bkt_width, label_font);
	bktwidth_visual_obj.setFont(label_font);
	bktwidth_visual_obj.setPenColor(Color.black);
	ZBounds bktwidth_bounds = bktwidth_visual_obj.getBounds();
	if(bktwidth_bounds.getWidth() < bounds_rect_.getWidth() &&
	   ((top_boundary+bktwidth_bounds.getHeight()+SPACING_X/camera_mag_) < 
	    (bounds_rect_.getY()+bounds_rect_.getWidth()))) {
	    Point2D bktwidth_location = new Point2D.Double(bounds_rect_.getX()+bounds_rect_.getWidth()/2.0
							   -bktwidth_bounds.getWidth()/2.0f, top_boundary);
	    bktwidth_visual_obj.setTranslation(bktwidth_location);
	    visual_components.addElement(bktwidth_visual_obj);
	}
	return visual_components;
    }
}

